@@ -14,17 +14,17 @@
 #    under the License.
 
 import pecan
-from pecan import rest
 from six.moves import http_client
 import wsme
 from wsme import types as wtypes
 
+from oslo_serialization import jsonutils
+
+from cyborg.agent.rpcapi import AgentAPI
 from cyborg.api.controllers import base
 from cyborg.api.controllers import link
-from cyborg.api.controllers.v1 import types
-from cyborg.api.controllers.v1 import utils as api_utils
+from cyborg.api.controllers import types
 from cyborg.api import expose
-from cyborg.common import exception
 from cyborg.common import policy
 from cyborg import objects
 
@@ -40,41 +40,23 @@ class Deployable(base.APIBase):
     uuid = types.uuid
     """The UUID of the deployable"""
 
+    parent_id = types.integer
+    """The parent ID of the deployable"""
+
+    root_id = types.integer
+    """The root ID of the deployable"""
+
     name = wtypes.text
     """The name of the deployable"""
 
-    parent_uuid = types.uuid
-    """The parent UUID of the deployable"""
+    num_accelerators = types.integer
+    """The number of accelerators of the deployable"""
 
-    root_uuid = types.uuid
-    """The root UUID of the deployable"""
+    device_id = types.integer
+    """The device on which the deployable is located"""
 
-    pcie_address = wtypes.text
-    """The pcie address of the deployable"""
-
-    host = wtypes.text
-    """The host on which the deployable is located"""
-
-    board = wtypes.text
-    """The board of the deployable"""
-
-    vendor = wtypes.text
-    """The vendor of the deployable"""
-
-    version = wtypes.text
-    """The version of the deployable"""
-
-    type = wtypes.text
-    """The type of the deployable"""
-
-    assignable = types.boolean
-    """Whether the deployable is assignable"""
-
-    instance_uuid = types.uuid
-    """The UUID of the instance which deployable is assigned to"""
-
-    availability = wtypes.text
-    """The availability of the deployable"""
+    attributes_list = wtypes.text
+    """The json list of attributes of the deployable"""
 
     links = wsme.wsattr([link.Link], readonly=True)
     """A list containing a self link"""
@@ -95,6 +77,13 @@ class Deployable(base.APIBase):
             link.Link.make_link('bookmark', url, 'deployables', api_dep.uuid,
                                 bookmark=True)
             ]
+        query = {"deployable_id": obj_dep.id}
+        attr_get_list = objects.Attribute.get_by_filter(pecan.request.context,
+                                                        query)
+        attributes_list = []
+        for exist_attr in attr_get_list:
+            attributes_list.append({exist_attr.key: exist_attr.value})
+        api_dep.attributes_list = jsonutils.dumps(attributes_list)
         return api_dep
 
 
@@ -119,12 +108,34 @@ class DeployablePatchType(types.JsonPatchType):
     @staticmethod
     def internal_attrs():
         defaults = types.JsonPatchType.internal_attrs()
-        return defaults + ['/pcie_address', '/host', '/type']
+        return defaults + ['/name', '/num_accelerators']
 
 
-class DeployablesController(rest.RestController):
+class DeployablesController(base.CyborgController):
     """REST controller for Deployables."""
 
+    _custom_actions = {'program': ['PATCH']}
+
+    @policy.authorize_wsgi("cyborg:deployable", "program", False)
+    @expose.expose(Deployable, types.uuid, body=[DeployablePatchType])
+    def program(self, uuid, program_info):
+        """Program a new deployable(FPGA).
+
+        :param uuid: The uuid of the target deployable.
+        :param program_info: JSON string containing what to program.
+        """
+
+        image_uuid = program_info[0]['value'][0]['image_uuid']
+        agent_api = AgentAPI()
+        obj_dep = objects.Deployable.get(pecan.request.context, uuid)
+        # Set attribute of the new bitstream/image information
+        obj_dep.add_attribute(pecan.request.context, 'image_uuid', image_uuid)
+        # TODO(Li Liu) Trigger the program api in Agnet.
+        agent_api.program_fpga_with_bitstream(pecan.request.context,
+                                              uuid,
+                                              image_uuid)
+        return Deployable.convert_with_links(obj_dep)
+
     @policy.authorize_wsgi("cyborg:deployable", "create", False)
     @expose.expose(Deployable, body=types.jsontype,
                    status_code=http_client.CREATED)
@@ -154,33 +165,54 @@ class DeployablesController(rest.RestController):
 
     @policy.authorize_wsgi("cyborg:deployable", "get_all")
     @expose.expose(DeployableCollection, int, types.uuid, wtypes.text,
-                   wtypes.text, types.boolean)
-    def get_all(self):
+                   wtypes.text, wtypes.ArrayType(types.FilterType))
+    # TODO(wangzhh): Remove limit, marker, sort_key, sort_dir in next release.
+    # They are used to compatible with R release client.
+    def get_all(self, limit=None, marker=None, sort_key='id', sort_dir='asc',
+                filters=None):
         """Retrieve a list of deployables."""
-        obj_deps = objects.Deployable.list(pecan.request.context)
+        filters_dict = {}
+        self._generate_filters(limit, sort_key, sort_dir, filters_dict)
+        if filters:
+            for filter in filters:
+                filters_dict.update(filter.as_dict())
+        context = pecan.request.context
+        if marker:
+            marker_obj = objects.Deployable.get(context, marker)
+            filters_dict["marker_obj"] = marker_obj
+        obj_deps = objects.Deployable.list(context, filters=filters_dict)
         return DeployableCollection.convert_with_links(obj_deps)
 
+    def _generate_filters(self, limit, sort_key, sort_dir, filters_dict):
+        """This method are used to compatible with R release client."""
+        if limit:
+            filters_dict["limit"] = limit
+        if sort_key:
+            filters_dict["sort_key"] = sort_key
+        if sort_dir:
+            filters_dict["sort_dir"] = sort_dir
+
     @policy.authorize_wsgi("cyborg:deployable", "update")
     @expose.expose(Deployable, types.uuid, body=[DeployablePatchType])
     def patch(self, uuid, patch):
         """Update a deployable.
 
+        Usage: curl -X PATCH {ip}:{port}/v1/accelerators/deployables/
+        {deployable_uuid} -d '[{"path":"/instance_uuid","value":
+        {instance_uuid}, "op":"replace"}]'  -H "Content-type:
+        application/json"
+
         :param uuid: UUID of a deployable.
         :param patch: a json PATCH document to apply to this deployable.
         """
         context = pecan.request.context
-        obj_dep = objects.Deployable.get(context, uuid)
 
-        try:
-            api_dep = Deployable(
-                **api_utils.apply_jsonpatch(obj_dep.as_dict(), patch))
-        except api_utils.JSONPATCH_EXCEPTIONS as e:
-            raise exception.PatchError(patch=patch, reason=e)
+        obj_dep = objects.Deployable.get(context, uuid)
 
         # Update only the fields that have changed
         for field in objects.Deployable.fields:
             try:
-                patch_val = getattr(api_dep, field)
+                patch_val = getattr(obj_dep, field)
             except AttributeError:
                 # Ignore fields that aren't exposed in the API
                 continue
